/*
	HereTic INFO.c processor, by Frans P. de Vries.

Derived from:

	DooM MaP StaTistics, by Frans P. de Vries.

	You are allowed to use any parts of this code in another program, as
	long as you give credits to the authors in the documentation and in
	the program itself.  Read the file README for more information.

	This program comes with absolutely no warranty.

	PRINT.C - Print routines
*/

/* the includes */
#include "htinfo.h"
#include "doomdef.h"

/* the local data */
#define MAXFRM 40
int frames[MAXFRM];
#define MAXSPAWN 10
int spawns[MAXSPAWN] = {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1};

typedef struct monstdata_s {
	int objt;
	int pain;
	int hlth;
} monstdata_t;

monstdata_t **monstdata = NULL;
int nummonst = 0;

/* the local functions */
void PrintSound(char *, int);
void PrintSpawns( void);
void PrintWpnAmmo( weapontype_t);
void PrintShotsPM( mobjinfo_t *, Bool);
void PrintWpnShotsPM( weapontype_t, Bool);
void PrintSpeed( mobjinfo_t *);
int FrameTics( statenum_t);
void PrintSprites( mobjinfo_t *);
void PrintWpnSprites( weapontype_t);
void PrintSprite( statenum_t, char *);
void TallyFrames( statenum_t);
int SpriteCount( void);
char *SpriteFrames( int);
float PainPerc( int);
int PainTics( statenum_t);


/*
	the weapon print data function
*/
void PrintWeapon()
{
	if (Weap < 0 || Weap >= NUMWEAPONS)
		ProgError( "no such Weapon %d", Weap);

	printf( "\n");
	printf( "Weapon\t\t%s\n", Wpn2Str( Weap));
	PrintWpnAmmo( Weap);
	PrintWpnShotsPM( Weap, FALSE);
	PrintWpnShotsPM( Weap, TRUE);
	printf( "\n");
	PrintWpnSprites( Weap);
	PrintSpawns();
}

/*
	the main print data function
*/
void PrintData()
{
	mobjinfo_t *MOBJ;

	if (Enum != NULL)
		Objt = Enum2Obj( Enum);
	else if (Thng != -1)
		Objt = Thg2Obj( Thng);
	else
		if (Objt < 0 || Objt >= NUMMOBJTYPES)
			ProgError( "no such Object %d", Objt);

	MOBJ = &mobjinfo[Objt];
	printf( "\n");
	printf( "Enum\t\t%s (%d)\n", Obj2Str( Objt), Objt);
	if (MOBJ->doomednum != -1)
		printf( "Type\t\t%d\t%X\n", MOBJ->doomednum, MOBJ->doomednum);
	else
		printf( "Type\t\t%d\n", MOBJ->doomednum);

	printf( "Health\t\t%d\n", MOBJ->spawnhealth);
	if (MOBJ->flags & MF_SPECIAL) // pick-up item
	{
		PrintShotsPM( MOBJ, FALSE);
		PrintShotsPM( MOBJ, TRUE);
	}
	else
		PrintSpeed( MOBJ);

	//printf( "Width\t\t%d\n", MOBJ->radius * 2 / FRACUNIT);
	printf( "Radius\t\t%d\n", MOBJ->radius / FRACUNIT);
	printf( "Height\t\t%d\n", MOBJ->height / FRACUNIT);
	printf( "Mass\t\t%d\n", MOBJ->mass);
	if (MOBJ->flags & MF_MISSILE)
		printf( "Damage\t\t%d-%d\n", MOBJ->damage, MOBJ->damage * 8);
	else
		printf( "Damage\t\t%d\n", MOBJ->damage);

	if (!(MOBJ->flags & MF_SPECIAL)) // no pick-up item
	{
		printf( "Reaction time\t%d tics\n", MOBJ->reactiontime);
		printf( "Pain chance\t%d ; %.2f%%\n", MOBJ->painchance, PainPerc( MOBJ->painchance));
		printf( "Pain time\t%d tics\n", PainTics( MOBJ->painstate));
	}
	printf( "Flags\t\t%08x  %d\n", MOBJ->flags, MOBJ->flags);
	PrintFlags( MOBJ->flags);
	printf( "Flags2\t\t%08x  %d\n", MOBJ->flags2, MOBJ->flags2);
	PrintFlags2( MOBJ->flags2);

//printf("spawnstate\n");
	TallyFrames( MOBJ->spawnstate);
//printf("seestate\n");
	TallyFrames( MOBJ->seestate);
//printf("painstate\n");
	TallyFrames( MOBJ->painstate);
//printf("meleestate\n");
	TallyFrames( MOBJ->meleestate);
//printf("missilestate\n");
	TallyFrames( MOBJ->missilestate);
//printf("crashstate\n");
	TallyFrames( MOBJ->crashstate);
//printf("deathstate\n");
	TallyFrames( MOBJ->deathstate);
//printf("xdeathstate\n");
	TallyFrames( MOBJ->xdeathstate);

	printf( "\n");
	if (MOBJ->flags & MF_SPECIAL) // pick-up item
		PrintSprites( MOBJ);
	else
	{
		PrintSprite( MOBJ->spawnstate, NULL);
		PrintSound( "Alert", MOBJ->seesound);
		PrintSound( "Active", MOBJ->activesound);
		PrintSound( "Attack", MOBJ->attacksound);
		PrintSound( "Pain", MOBJ->painsound);
		PrintSound( "Death", MOBJ->deathsound);
	}
	PrintSpawns();
}


void PrintSound(char *type, int sound)
{
	char name[9];

	printf( "%s sound\t", type);
	if (sound != 0)
	{
		if (S_sfx[sound].link != NULL)
			strcpy( name, S_sfx[sound].link->name);
		else
			strcpy( name, S_sfx[sound].name);
		strupr( name);
		printf( "%s\n", name);
	}
	else
		printf( "-\n");
}

void PrintSpawns( void)
{
	int i;

	for (i = 0; i < MAXSPAWN; i++)
		if (spawns[i] != -1)
		{
			if (i == 0)
				printf( "\nSpawns\t");
			else
				printf( "\n\t");
			printf( "\t%s", Obj2Str( spawns[i]));
		}
	printf( "\n");
}

void PrintWpnAmmo( weapontype_t weapon)
{
	int ammo;

	ammo = GetWeaponAmmo[weapon];
	printf("Incl ammo\t%d (%d on [[Skill level|skills]] 1 & 5)\n", ammo, (int)(ammo * 1.5));
	ammo = maxammo[wpnlev1info[weapon].ammo];
	printf("Max ammo\t%d (%d with [[bag of holding]])\n", ammo, ammo * 2);
}


void PrintShotsPM( mobjinfo_t *MOBJ, Bool powered)
{
	int weapon;

	switch (states[MOBJ->spawnstate].sprite)
	{
		case SPR_WGNT: weapon = wp_gauntlets; break;
		case SPR_WBOW: weapon = wp_crossbow; break;
		case SPR_WBLS: weapon = wp_blaster; break;
		case SPR_WSKL: weapon = wp_skullrod; break;
		case SPR_WPHX: weapon = wp_phoenixrod; break;
		case SPR_WMCE: weapon = wp_mace; break;
		default: return;
	}
	PrintWpnShotsPM( weapon, powered);
}

void PrintWpnShotsPM( weapontype_t weapon, Bool powered)
{
	int percycle = 0, tics = 0;
	statenum_t state, st;

	state = st = (powered ? wpnlev2info[weapon].holdatkstate
	                      : wpnlev1info[weapon].holdatkstate);
	if (state != S_NULL)
		for (;;)
		{
			if (states[st].action)
				states[st].action(NULL);
			else
				funcname[0] = '\0';
			if (strncmp( funcname, "A_Fire", strlen("A_Fire")) == 0 ||
			    strncmp( funcname, "A_StaffAttack", strlen("A_StaffAttack")) == 0 ||
			    strncmp( funcname, "A_BeakAttack", strlen("A_BeakAttack")) == 0 ||
			    strcmp( funcname, "A_GauntletAttack") == 0)
				percycle++;
			else if (strcmp( funcname, "A_ReFire") == 0 ||
			         strcmp( funcname, "A_BeakReady") == 0)
				break;
			tics += states[st].tics;
			st = states[st].nextstate;
			if (tics > 100)
				ProgError( "unable to count tics for holdatkstate %d", state);
		}

	if (tics != 0)
		printf( "Shots / min%s\t%.1f\n", (powered ? " (p)" : ""),
		        (float)percycle * (TICRATE * 60) / tics);
}

void PrintSpeed( mobjinfo_t *MOBJ)
{
	int tics;

	if (MOBJ->flags & MF_MISSILE)
		printf( "Speed\t\t%d / tic ; %d / sec\n",
		        MOBJ->speed / FRACUNIT, MOBJ->speed / FRACUNIT * 35);
	else
	{
		tics = FrameTics( MOBJ->seestate);
		if (tics != -1)
			printf( "Speed\t\t%d / frame ; %.1f / sec\n",
			        MOBJ->speed, MOBJ->speed * 35.0f / tics);
		else
			printf( "Speed\t\t%d\n", MOBJ->speed);
	}
}

int FrameTics( statenum_t state)
{
	if (states[state].action)
	{
		states[state].action(NULL);
		if (strcmp( funcname, "A_Chase") == 0 ||
		    strcmp( funcname, "A_ChicChase") == 0 ||  // Chicken
		    strcmp( funcname, "A_Sor1Chase") == 0)    // D'Sparil
			return states[state].tics;
	}

	return -1;
}


void PrintSprites( mobjinfo_t *MOBJ)
{
	int weapon;

	PrintSprite( MOBJ->spawnstate, "pickup");

	switch (states[MOBJ->spawnstate].sprite)
	{
		case SPR_WGNT: weapon = wp_gauntlets; break;
		case SPR_WBOW: weapon = wp_crossbow; break;
		case SPR_WBLS: weapon = wp_blaster; break;
		case SPR_WSKL: weapon = wp_skullrod; break;
		case SPR_WPHX: weapon = wp_phoenixrod; break;
		case SPR_WMCE: weapon = wp_mace; break;
		default: return;
	}
	PrintWpnSprites( weapon);
}

void PrintWpnSprites( weapontype_t weapon)
{
	PrintSprite( wpnlev1info[weapon].readystate, "wield");
	PrintSprite( wpnlev1info[weapon].atkstate, "fire");
	PrintSprite( wpnlev1info[weapon].holdatkstate, "refire");
	PrintSprite( wpnlev2info[weapon].readystate, "p wld");
	PrintSprite( wpnlev2info[weapon].atkstate, "p fire");
	PrintSprite( wpnlev2info[weapon].holdatkstate, "p ref");
}

void PrintSprite( statenum_t state, char *type)
{
	int num;
	char *frm;

	if (state == S_NULL)
		return;
	if (type != NULL)
		printf( "Sprite (%s)\t%s\n", type, sprnames[states[state].sprite]);
	else
		printf( "Sprite name\t%s\n", sprnames[states[state].sprite]);
	TallyFrames( state);
	num = SpriteCount();
	frm = SpriteFrames( num);
	printf( "Frames\t\t%d (%s)\n", num, frm);
	FreeMemory( frm);
}

void TallyFrames( statenum_t state)
{
	statenum_t st = state;
	int f, num = 0;

	// reset seen frames
	for (f = 0; f < MAXFRM; f++)
		frames[f] = -1;
	// ignore null state
	if (state == S_NULL)
		return;

	for (;;)
	{
		// tally this frame
		f = states[st].frame & FF_FRAMEMASK;
		if (f >= MAXFRM)
			ProgError( "unable to count frames for state %d", state);
		frames[num++] = f;

		// done if no next state, or same state, or looped around to earlier one
		if (states[st].nextstate == S_NULL || states[st].nextstate <= st)
			break;
		st = states[st].nextstate;
		// done if looped around
		if (st == state)
			break;

		// cease flashing
		if (st == S_LIGHTDONE)
			break;
		if (states[st].action)
			states[st].action(NULL);
		else
			funcname[0] = '\0';
//printf("%s\n", funcname);
		// cease firing
		if (strcmp( funcname, "A_ReFire") == 0 ||
		    strcmp( funcname, "A_VolcanoSet") == 0 ||     // S_VOLCANO9 loops to S_VOLCANO2, not 1
		    strcmp( funcname, "A_GenWizard") == 0 ||      // S_SOR2FX2_3 loops to S_SOR2FX2_2, not 1
		    strcmp( funcname, "A_WhirlwindSeek") == 0 ||  // S_HEADFX4_7 loops to S_HEADFX4_5, not 1
		    strcmp( funcname, "A_BeakReady") == 0)        // S_BEAKREADY loops to S_BEAKREADY
			break;
	}
}

int SpriteCount( void)
{
	int f, count = 0;

	for (f = 0; f < MAXFRM && frames[f] != -1; f++)
		count++;

	return count;
}

char *SpriteFrames( int count)
{
	char *ret, frm[MAXFRM];
	int f;

	memset(frm, 0, MAXFRM);
	// transpose char to actually used frame
	for (f = 0; f < MAXFRM && frames[f] != -1; f++)
		frm[f] = 'A' + frames[f];

	return strdup( frm);
}


float PainPerc( int chance)
{
	int i, count = 0;

	for (i = 0; i < 256; i++)
		if (rndtable[i] < chance)
			count++;

	// ensure x.xx50 is rounded up to x.xx+0.01
	return count * 100.0f / 256.0f + 0.0001f;
}

void PainChances( void)
{
	int i;

	printf( "\n");
	for (i = 0; i <= 256; i++)
		printf( "%3d\t%.2f%%\n", i, PainPerc( i));
}

int PainTics( statenum_t state)
{
	statenum_t st = state;
	int tics = 0;

	if (st != S_NULL)
		for (;;)
		{
			tics += states[st].tics;
			if (states[st].action)
				states[st].action(NULL);
			else
				funcname[0] = '\0';
			if (strcmp( funcname, "A_Pain") == 0 ||
			    strcmp( funcname, "A_PodPain") == 0 ||
			    strcmp( funcname, "A_ChicPain") == 0 ||
			    strcmp( funcname, "A_Sor1Pain") == 0)
				break;
			st = states[st].nextstate;
			if (tics > 30)
				ProgError( "unable to count tics for painstate %d", state);
		}

	return tics;
}


/* sort by decreasing pain chaince, decreasing hit points, alphabetic name */
int SortMonst( const void *m1, const void *m2)
{
	monstdata_t *p1 = * (monstdata_t * const *) m1,
	            *p2 = * (monstdata_t * const *) m2;

	if (p1->pain < p2->pain)
		return 1;
	else if (p1->pain > p2->pain)
		return -1;

	if (p1->hlth < p2->hlth)
		return 1;
	else if (p1->hlth > p2->hlth)
		return -1;
	
	return strcmp( Monst2Str( p1->objt),
	               Monst2Str( p2->objt));
}

void MonstersPcHp( void)
{
	mobjinfo_t *MOBJ;
	monstdata_t *mnst;

	// skip player
	for (Objt = 1; Objt < NUMMOBJTYPES; Objt++)
		if (Objt != MT_PLAYER && Objt != MT_CHICPLAYER) {
			MOBJ = &mobjinfo[Objt];
			if (MOBJ->painchance != 0)
			{
				mnst = (monstdata_t *) GetMemory( sizeof(monstdata_t));
				mnst->objt = Objt;
				mnst->pain = MOBJ->painchance;
				mnst->hlth = MOBJ->spawnhealth;
				monstdata = ResizeMemory( monstdata, (nummonst+1) * sizeof(mnst));
				monstdata[nummonst++] = mnst;
			}
		}

	qsort( monstdata, nummonst, sizeof(monstdata_t *), SortMonst);

	for (Objt = 0; Objt < nummonst; Objt++)
	{
		printf( "%-20s\t%d (%.2f%%)\t%d\n",
		        Monst2Str( monstdata[Objt]->objt),
		        monstdata[Objt]->pain, PainPerc( monstdata[Objt]->pain),
		        monstdata[Objt]->hlth);
	}
}

/* vim:set noexpandtab tabstop=2 softtabstop=2 shiftwidth=2: */
