/*
	HeXen INFO.c processor, by Frans P. de Vries.

Derived from:

	DooM MaP StaTistics, by Frans P. de Vries.

	You are allowed to use any parts of this code in another program, as
	long as you give credits to the authors in the documentation and in
	the program itself.  Read the file README for more information.

	This program comes with absolutely no warranty.

	HXINFO.H - Main HXINFO defines.
*/


/* the includes */
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "dmunix.h"
#include "sounds.h"

/* global definitions */
#define HXINFO_VERSION "1.0"

typedef int16_t  BCINT;
typedef uint16_t UBCINT;
typedef uint8_t  UBYTE;
typedef int32_t  BCLNG;


/*
	syntactic sugar
*/
typedef BCINT Bool; /* Boolean data: true or false */

/* boolean constants */
#ifndef TRUE
#define TRUE  1
#define FALSE 0
#endif

/*
	Description of the command line arguments and config file keywords
*/
typedef struct
{
	char *short_name;   /* abbreviated command line argument */
	char *long_name;    /* command line arg. or keyword */
	enum {              /* type of this option: */
		OPT_BOOLEAN,      /* boolean (toggle) */
		OPT_INTEGER,      /* integer number */
		OPT_STRING,       /* character string */
		OPT_STRINGACC,    /* character string, but store in a list */
		OPT_STRINGLIST,   /* list of character strings */
		OPT_END           /* end of the options description */
	}     opt_type;
	char *msg_if_true;  /* message printed if option is true */
	char *msg_if_false; /* message printed if option is false */
	void *data_ptr;     /* pointer to the data */
} OptDesc;


/*
	the interfile global variables
*/
extern sfxinfo_t S_sfx[];
//extern unsigned char rndtable[];
extern int WeaponManaUse[][NUMWEAPONS];
extern char funcname[];

extern int Thng;   /* thing number */
extern int Objt;   /* object number */
extern int Weap;   /* weapon number */
extern int Clss;   /* class number */
extern char *Enum; /* enum string */
extern Bool Pain;  /* pain chance table */
extern Bool Mnst;  /* monsters pc/hp */
extern Bool AllO;  /* all objects */

/*
	the function prototypes
*/

/* from hxinfo.c */
int main( int, char *[]);
void ParseCommandLineOptions( int, char *[]);
void Usage( FILE *);
void Credits( FILE *);
void ProgError( char *, ...);

/* from print.c */
void PrintWeapon( void);
void PrintData( void);
void PainChances( void);
void MonstersPcHp( void);

/* from maps.c */
int Thg2Obj( int);
char *Obj2Str( int);
char *Wpn2Str( int, int);
int Enum2Obj( char *);
void PrintFlags( int);
char *Monst2Str( int);

/* vim:set noexpandtab tabstop=2 softtabstop=2 shiftwidth=2: */
