/*
	HeXen INFO.c processor, by Frans P. de Vries.

Derived from:

	DooM MaP StaTistics, by Frans P. de Vries.

	You are allowed to use any parts of this code in another program, as
	long as you give credits to the authors in the documentation and in
	the program itself.  Read the file README for more information.

	This program comes with absolutely no warranty.

	PRINT.C - Print routines
*/

/* the includes */
#include "hxinfo.h"
#include "h2def.h"

/* the local data */
#define MAXFRM 40
int frames[MAXFRM];
#define MAXSPAWN 10
int spawns[MAXSPAWN] = {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1};

typedef struct monstdata_s {
	int objt;
	int pain;
	int hlth;
} monstdata_t;

monstdata_t **monstdata = NULL;
int nummonst = 0;

/* the local functions */
void PrintSound(char *, int);
char *LookupSoundTag( char *);
void PrintSpawns( void);
void PrintWpnMana( weapontype_t, pclass_t);
void PrintShotsPM( mobjinfo_t *);
void PrintWpnShotsPM( weapontype_t, pclass_t);
void PrintSpeed( mobjinfo_t *);
int FrameTics( statenum_t);
void PrintSprites( mobjinfo_t *);
void PrintWpnSprites( weapontype_t, pclass_t);
void PrintSprite( statenum_t, char *);
void TallyFrames( statenum_t);
int SpriteCount( void);
char *SpriteFrames( int);
float PainPerc( int);
int PainTics( statenum_t);


/*
	the weapon print data function
*/
void PrintWeapon()
{
	if (Weap < 0 || Weap >= NUMWEAPONS)
		ProgError( "no such Weapon %d", Weap);
	if (Clss < 0 || Clss >= NUMCLASSES)
		ProgError( "no such Class %d", Clss);

	printf( "\n");
	printf( "Weapon\t\t%s\n", Wpn2Str( Weap, Clss));
	PrintWpnMana( Weap, Clss);
	printf( "\n");
	PrintWpnShotsPM( Weap, Clss);
	PrintWpnSprites( Weap, Clss);
	PrintSpawns();
}

/*
	the main print data function
*/
void PrintData()
{
	mobjinfo_t *MOBJ;

	if (Enum != NULL)
		Objt = Enum2Obj( Enum);
	else if (Thng != -1)
		Objt = Thg2Obj( Thng);
	else
		if (Objt < 0 || Objt >= NUMMOBJTYPES)
			ProgError( "no such Object %d", Objt);

	MOBJ = &mobjinfo[Objt];
	printf( "\n");
	printf( "Enum\t\t%s (%d)\n", Obj2Str( Objt), Objt);
	if (MOBJ->doomednum != -1)
		printf( "Type\t\t%d\t%X\n", MOBJ->doomednum, MOBJ->doomednum);
	else
		printf( "Type\t\t%d\n", MOBJ->doomednum);

	printf( "Health\t\t%d\n", MOBJ->spawnhealth);
	if (MOBJ->flags & MF_SPECIAL) // pick-up item
		PrintShotsPM( MOBJ);
	else
		PrintSpeed( MOBJ);

	//printf( "Width\t\t%d\n", MOBJ->radius * 2 / FRACUNIT);
	printf( "Radius\t\t%d\n", MOBJ->radius / FRACUNIT);
	printf( "Height\t\t%d\n", MOBJ->height / FRACUNIT);
	printf( "Mass\t\t%d\n", MOBJ->mass);
	if (MOBJ->flags & MF_MISSILE)
		printf( "Damage\t\t%d-%d\n", MOBJ->damage, MOBJ->damage * 8);
	else
		printf( "Damage\t\t%d\n", MOBJ->damage);

	if (!(MOBJ->flags & MF_SPECIAL)) // no pick-up item
	{
		printf( "Reaction time\t%d tics\n", MOBJ->reactiontime);
		printf( "Pain chance\t%d ; %.2f%%\n", MOBJ->painchance, PainPerc( MOBJ->painchance));
		printf( "Pain time\t%d tics\n", PainTics( MOBJ->painstate));
	}
	printf( "Flags\t\t%08x  %d\n", MOBJ->flags, MOBJ->flags);
	PrintFlags( MOBJ->flags);
	printf( "Flags2\t\t%08x  %d\n", MOBJ->flags2, MOBJ->flags2);
	PrintFlags2( MOBJ->flags2);

	TallyFrames( MOBJ->spawnstate);
	TallyFrames( MOBJ->seestate);
	TallyFrames( MOBJ->painstate);
	TallyFrames( MOBJ->meleestate);
	TallyFrames( MOBJ->missilestate);
	TallyFrames( MOBJ->crashstate);
	TallyFrames( MOBJ->deathstate);
	TallyFrames( MOBJ->xdeathstate);

	printf( "\n");
	if (MOBJ->flags & MF_SPECIAL) // pick-up item
		PrintSprites( MOBJ);
	else
	{
		PrintSprite( MOBJ->spawnstate, NULL);
		PrintSound( "Alert", MOBJ->seesound);
		PrintSound( "Active", MOBJ->activesound);
		PrintSound( "Attack", MOBJ->attacksound);
		PrintSound( "Pain", MOBJ->painsound);
		PrintSound( "Death", MOBJ->deathsound);
	}
	PrintSpawns();
}


void PrintSound(char *type, int sound)
{
	char *name;

	printf( "%s sound\t", type);

	if (sound != 0)
	{
		name = LookupSoundTag( S_sfx[sound].tagName);
		strupr( name);
		printf( "%s\n", name);
		FreeMemory( name);
	}
	else
		printf( "-\n");
}

#define BUFSIZE 120

char *LookupSoundTag( char *tag)
{
	FILE *sndinfo;
	char buf[BUFSIZE], *ptr = NULL;

	if ((sndinfo = fopen( "sndinfo.txt", "r")) == NULL)
		ProgError( "error opening sndinfo.txt");

	while (fgets( buf, BUFSIZE, sndinfo) != NULL)
	{
		/* skip comment lines */
		if (buf[0] == ';')
			continue;

		if (strncmp(buf, tag, strlen(tag)) == 0)
		{
			ptr = strrchr( buf, ' ') + 1;
			/* erase newline */
			buf[strlen(buf)-1] = '\0';
			break;
		}
	}

	fclose( sndinfo); 
	if (ptr != NULL)
		return strdup( ptr);
	else
		return strdup( "");
}

void PrintSpawns( void)
{
	int i;

	for (i = 0; i < MAXSPAWN; i++)
		if (spawns[i] != -1)
		{
			if (i == 0)
				printf( "\nSpawns\t");
			else
				printf( "\n\t");
			printf( "\t%s", Obj2Str( spawns[i]));
		}
	printf( "\n");
}

char *ManaColor( weapontype_t weapon, pclass_t class)
{
	if (WeaponInfo[weapon][class].mana == MANA_1)
		return "Blue";
	else if (WeaponInfo[weapon][class].mana == MANA_2)
		return "Green";
	else if (WeaponInfo[weapon][class].mana == MANA_BOTH)
		return "Both";
	else
		return "None";
}

void PrintWpnMana( weapontype_t weapon, pclass_t class)
{
	int mana;

	if (weapon == WP_FIRST)
	{
		printf("Incl mana\tInfinite\n");
		return;
	}
	mana = (weapon == WP_FOURTH) ? 20 : 25; // hardcoded
	printf("Incl mana\t%d %s (%d on [[Skill level|skills]] 1 & 5)\n",
	       mana, ManaColor( weapon, class), (int)(mana * 1.5));
	mana = WeaponManaUse[class][weapon];
	printf("Mana use\t%d %s\n", mana, ManaColor( weapon, class));
}


void PrintShotsPM( mobjinfo_t *MOBJ)
{
	int weapon, class;

	switch (states[MOBJ->spawnstate].sprite)
	{
		case SPR_WFAX: weapon = WP_SECOND; class = PCLASS_FIGHTER; break;
		case SPR_WCSS: weapon = WP_SECOND; class = PCLASS_CLERIC; break;
		case SPR_WMCS: weapon = WP_SECOND; class = PCLASS_MAGE; break;
		case SPR_WFHM: weapon = WP_THIRD; class = PCLASS_FIGHTER; break;
		case SPR_WCFM: weapon = WP_THIRD; class = PCLASS_CLERIC; break;
		case SPR_WMLG: weapon = WP_THIRD; class = PCLASS_MAGE; break;
		case SPR_WFR1:
		case SPR_WFR2:
		case SPR_WFR3: weapon = WP_FOURTH; class = PCLASS_FIGHTER; break;
		case SPR_WCH1:
		case SPR_WCH2:
		case SPR_WCH3: weapon = WP_FOURTH; class = PCLASS_CLERIC; break;
		case SPR_WMS1:
		case SPR_WMS2:
		case SPR_WMS3: weapon = WP_FOURTH; class = PCLASS_MAGE; break;
		default: return;
	}
	PrintWpnShotsPM( weapon, class);
}

void PrintWpnShotsPM( weapontype_t weapon, pclass_t class)
{
	int percycle = 0, tics = 0;
	statenum_t state, st;

	state = st = WeaponInfo[weapon][class].holdatkstate;
	if (state != S_NULL)
		for (;;)
		{
			if (states[st].action)
				states[st].action(NULL);
			else
				funcname[0] = '\0';
			if (strncmp( funcname, "A_Fire", strlen("A_Fire")) == 0 ||
			    strcmp( funcname, "A_MntrFloorFire") == 0)
				percycle++;
			else if (strcmp( funcname, "A_ReFire") == 0 ||
			         strcmp( funcname, "A_CStaffAttack") == 0 ||      // Serpent Staff
			         strcmp( funcname, "A_FHammerThrow") == 0 ||      // Hammer of Retribution
			         strcmp( funcname, "A_CFlameAttack") == 0 ||      // Firestorm
			         strcmp( funcname, "A_MLightningAttack") == 0 ||  // Arc of Death
			         strcmp( funcname, "A_FSwordAttack") == 0 ||      // Quietus
			         strcmp( funcname, "A_CHolyAttack") == 0 ||       // Wraithverge
			         strcmp( funcname, "A_MStaffAttack") == 0)        // Bloodscourge
				break;
			tics += states[st].tics;
			st = states[st].nextstate;
			if (tics > 100)
				ProgError( "unable to count tics for holdatkstate %d", state);
		}

	if (tics != 0)
		printf( "Shots / min\t%.1f\n",
		        (float)percycle * (TICRATE * 60) / tics);
}

void PrintSpeed( mobjinfo_t *MOBJ)
{
	int tics;

	if (MOBJ->flags & MF_MISSILE)
		printf( "Speed\t\t%d / frame ; %d / sec\n",
		        MOBJ->speed / FRACUNIT, MOBJ->speed / FRACUNIT * 35);
	else
	{
		tics = FrameTics( MOBJ->seestate);
		if (tics != -1)
			printf( "Speed\t\t%d / tic ; %.1f / sec\n",
			        MOBJ->speed, MOBJ->speed * 35.0f / tics);
		else
			printf( "Speed\t\t%d\n", MOBJ->speed);
	}
}

int FrameTics( statenum_t state)
{
	if (states[state].action)
	{
		states[state].action(NULL);
		if (strcmp( funcname, "A_Chase") == 0 ||
		    strcmp( funcname, "A_PigChase") == 0 ||       // Chicken
		    strcmp( funcname, "A_WraithChase") == 0 ||    // Wraith
		    strcmp( funcname, "A_MinotaurChase") == 0 ||  // Maulotaur
		    strcmp( funcname, "A_SerpentChase") == 0 ||   // Serpent
		    strcmp( funcname, "A_BishopChase") == 0 ||    // Dark Bishop
		    strcmp( funcname, "A_FiredChase") == 0 ||     // Firedemon
		    strcmp( funcname, "A_IceGuyChase") == 0 ||    // Chicken
		    strcmp( funcname, "A_FastChase") == 0 ||      // Chicken
		    strcmp( funcname, "A_KoraxChase") == 0)       // Korax
		{
//printf("%s\n", funcname);
			return states[state].tics;
		}
	}

	return -1;
}


void PrintSprites( mobjinfo_t *MOBJ)
{
	int weapon, class;

	PrintSprite( MOBJ->spawnstate, "pickup");

	switch (states[MOBJ->spawnstate].sprite)
	{
		case SPR_WFAX: weapon = WP_SECOND; class = PCLASS_FIGHTER; break;
		case SPR_WCSS: weapon = WP_SECOND; class = PCLASS_CLERIC; break;
		case SPR_WMCS: weapon = WP_SECOND; class = PCLASS_MAGE; break;
		case SPR_WFHM: weapon = WP_THIRD; class = PCLASS_FIGHTER; break;
		case SPR_WCFM: weapon = WP_THIRD; class = PCLASS_CLERIC; break;
		case SPR_WMLG: weapon = WP_THIRD; class = PCLASS_MAGE; break;
		case SPR_WFR1:
		case SPR_WFR2:
		case SPR_WFR3: weapon = WP_FOURTH; class = PCLASS_FIGHTER; break;
		case SPR_WCH1:
		case SPR_WCH2:
		case SPR_WCH3: weapon = WP_FOURTH; class = PCLASS_CLERIC; break;
		case SPR_WMS1:
		case SPR_WMS2:
		case SPR_WMS3: weapon = WP_FOURTH; class = PCLASS_MAGE; break;
		default: return;
	}
	PrintWpnSprites( weapon, class);
}

void PrintWpnSprites( weapontype_t weapon, pclass_t class)
{
	PrintSprite( WeaponInfo[weapon][class].readystate, "wield");
	PrintSprite( WeaponInfo[weapon][class].atkstate, "fire");
	PrintSprite( WeaponInfo[weapon][class].holdatkstate, "refire");
}

void PrintSprite( statenum_t state, char *type)
{
	int num;
	char *frm;

	if (state == S_NULL)
		return;
	if (type != NULL)
		printf( "Sprite (%s)\t%s\n", type, sprnames[states[state].sprite]);
	else
		printf( "Sprite name\t%s\n", sprnames[states[state].sprite]);
	TallyFrames( state);
	num = SpriteCount();
	frm = SpriteFrames( num);
	printf( "Frames\t\t%d (%s)\n", num, frm);
	FreeMemory( frm);
}

void TallyFrames( statenum_t state)
{
	statenum_t st = state;
	int f, num = 0;

	// reset seen frames
	for (f = 0; f < MAXFRM; f++)
		frames[f] = -1;

	for (;;)
	{
		// tally this frame
		f = states[st].frame & FF_FRAMEMASK;
		if (f >= MAXFRM)
			ProgError( "unable to count frames for state %d", state);
		frames[num++] = f;

		// done if no next state, or same state, or looped around to earlier one
		if (states[st].nextstate == S_NULL || states[st].nextstate <= st)
			break;
		st = states[st].nextstate;
		// done if looped around to start state
		if (st == state)
			break;

		// cease flashing
		if (st == S_LIGHTDONE)
			break;
		if (states[st].action)
			states[st].action(NULL);
		else
			funcname[0] = '\0';
		// cease firing
		if (strcmp( funcname, "A_ReFire") == 0)
			break;
	}
}

int SpriteCount( void)
{
	int f, count = 0;

	for (f = 0; f < MAXFRM && frames[f] != -1; f++)
		count++;

	return count;
}

char *SpriteFrames( int count)
{
	char *ret, frm[MAXFRM];
	int f;

	memset(frm, 0, MAXFRM);
	// transpose char to actually used frame
	for (f = 0; f < MAXFRM && frames[f] != -1; f++)
		frm[f] = 'A' + frames[f];

	return strdup( frm);
}


float PainPerc( int chance)
{
	int i, count = 0;

	for (i = 0; i < 256; i++)
		if (rndtable[i] < chance)
			count++;

	// ensure x.xx50 is rounded up to x.xx+0.01
	return count * 100.0f / 256.0f + 0.0001f;
}

void PainChances( void)
{
	int i;

	printf( "\n");
	for (i = 0; i <= 256; i++)
		printf( "%3d\t%.2f%%\n", i, PainPerc( i));
}

int PainTics( statenum_t state)
{
	statenum_t st = state;
	int tics = 0;

	if (st != S_NULL)
		for (;;)
		{
			tics += states[st].tics;
			if (states[st].action)
				states[st].action(NULL);
			else
				funcname[0] = '\0';
			if (strcmp( funcname, "A_Pain") == 0 ||
			    strcmp( funcname, "A_PigPain") == 0 ||
			    strcmp( funcname, "A_DragonPain") == 0 ||
			    strcmp( funcname, "A_BishopPainBlur") == 0 ||
			    strcmp( funcname, "A_BounceCheck") == 0)
				break;
			st = states[st].nextstate;
			if (tics > 30)
				ProgError( "unable to count tics for painstate %d", state);
		}

	return tics;
}


/* sort by decreasing pain chaince, decreasing hit points, alphabetic name */
int SortMonst( const void *m1, const void *m2)
{
	monstdata_t *p1 = * (monstdata_t * const *) m1,
	            *p2 = * (monstdata_t * const *) m2;

	if (p1->pain < p2->pain)
		return 1;
	else if (p1->pain > p2->pain)
		return -1;

	if (p1->hlth < p2->hlth)
		return 1;
	else if (p1->hlth > p2->hlth)
		return -1;
	
	return strcmp( Monst2Str( p1->objt),
	               Monst2Str( p2->objt));
}

void MonstersPcHp( void)
{
	mobjinfo_t *MOBJ;
	monstdata_t *mnst;

	// skip players
	for (Objt = 1; Objt < NUMMOBJTYPES; Objt++)
		if (Objt != MT_PLAYER_FIGHTER && Objt != MT_PLAYER_CLERIC &&
		    Objt != MT_PLAYER_MAGE && Objt != MT_PIGPLAYER) {
			MOBJ = &mobjinfo[Objt];
			if (MOBJ->painchance != 0)
			{
				mnst = (monstdata_t *) GetMemory( sizeof(monstdata_t));
				mnst->objt = Objt;
				mnst->pain = MOBJ->painchance;
				mnst->hlth = MOBJ->spawnhealth;
				monstdata = ResizeMemory( monstdata, (nummonst+1) * sizeof(mnst));
				monstdata[nummonst++] = mnst;
			}
		}

	qsort( monstdata, nummonst, sizeof(monstdata_t *), SortMonst);

	for (Objt = 0; Objt < nummonst; Objt++)
	{
		printf( "%-20s\t%d (%.2f%%)\t%d\n",
		        Monst2Str( monstdata[Objt]->objt),
		        monstdata[Objt]->pain, PainPerc( monstdata[Objt]->pain),
		        monstdata[Objt]->hlth);
	}
}

/* vim:set noexpandtab tabstop=2 softtabstop=2 shiftwidth=2: */
