/*
	StriFe INFO.c processor, by Frans P. de Vries.

Derived from:

	DooM MaP StaTistics, by Frans P. de Vries.

	You are allowed to use any parts of this code in another program, as
	long as you give credits to the authors in the documentation and in
	the program itself.  Read the file README for more information.

	This program comes with absolutely no warranty.

	SFINFO.C - Main program routines
*/


/* the includes */
#include "sfinfo.h"
#include "info.h"


/* global variables */
int Thng = -1;     /* thing number */
int Objt = -1;     /* object number */
int Weap = -1;     /* weapon number */
char *Enum = NULL; /* enum string */
Bool Pain = FALSE; /* pain chance table */
Bool Mnst = FALSE; /* monsters pc/hp */
Bool AllO = FALSE; /* all objects */


OptDesc options[] =  /* description of the command line options */
{
/* short/long names  type  msg if true/changed  msg if false  store value */
	{ "T",  "THING",   OPT_INTEGER, "Thing number",  NULL, &Thng },
	{ "O",  "OBJECT",  OPT_INTEGER, "Object number", NULL, &Objt },
	{ "W",  "WEAPON",  OPT_INTEGER, "Weapon number", NULL, &Weap },
	{ "E",  "ENUM",    OPT_STRING,  "Enum string",   NULL, &Enum },
	{ "P",  "PAINTAB", OPT_BOOLEAN, "Pain chances",  NULL, &Pain },
	{ "M",  "MONSTERS",OPT_BOOLEAN, "Monsters PC/HP",NULL, &Mnst },
	{ "A",  "ALLOBJ",  OPT_BOOLEAN, "All objects",   NULL, &AllO },
	{ NULL, NULL,      OPT_END,     NULL,            NULL, NULL  }
};


/*
	the main program
*/
int main( int argc, char *argv[])
{
	Credits( stdout);
	argv++;
	argc--;
	/* read command line options */
	ParseCommandLineOptions( argc, argv);

	if (Pain)
	{
		PainChances();
		exit(0);
	}

	if (Mnst)
	{
		MonstersPcHp();
		exit(0);
	}

	if (AllO)
	{
		printf( "==================================================\n");
		for (Objt = 0; Objt < NUMMOBJTYPES; Objt++)
		{
			PrintData();
			printf( "==================================================\n");
		}
		exit(0);
	}

	if (Weap != -1)
	{
		PrintWeapon();
		exit(0);
	}

	if (Thng <= -1 && Objt <= -1 && Enum == NULL)
		ProgError( "no Thing or Object number or Enum string specified");
	if ((Thng != -1 && Objt != -1) ||
	    (Thng != -1 && Enum != NULL) ||
	    (Objt != -1 && Enum != NULL))
		ProgError( "must specify one of Thing number, Object number or Enum string");

	PrintData();
	exit( 0);
}


/*
	append a string to a null-terminated string list
*/
void AppendItemToList( char ***list, char *item)
{
	int i = 0;

	if (*list)
	{
		/* count the number of elements in the list (last = null) */
		while ((*list)[i])
			i++;
		/* expand the list */
		*list = (char **) ResizeMemory( *list, (i + 2) * sizeof( char **));
	}
	else
		/* create a new list */
		*list = (char **) GetMemory( 2 * sizeof( char **));

	/* append the new element */
	(*list)[i] = item;
	(*list)[i + 1] = NULL;
}

/*
	handle command line options
*/
void ParseCommandLineOptions( int argc, char *argv[])
{
	int optnum;

	while (argc > 0)
	{
		if (argv[0][0] != '-' && argv[0][0] != '+')
			ProgError( "options must start with '-' or '+'");
		strupr( argv[0]);
		if (!strcmp( argv[0], "-?") || !strcmp( argv[0], "-H") ||
		    !strcmp( argv[0], "-HELP"))
		{
			Usage( stdout);
			exit( 0);
		}
		for (optnum = 0; options[optnum].opt_type != OPT_END; optnum++)
		{
			if (!strcmp( &(argv[0][1]), options[optnum].short_name) ||
			    !strcmp( &(argv[0][1]), options[optnum].long_name))
			{
				switch (options[optnum].opt_type)
				{
				case OPT_BOOLEAN:
					if (argv[0][0] == '+') /* '+/-' reversed for print flags */
					{
						*((Bool *) (options[optnum].data_ptr)) = TRUE;
						if (options[optnum].msg_if_true)
							printf( "%s.\n", options[optnum].msg_if_true);
					}
					else
					{
						*((Bool *) (options[optnum].data_ptr)) = FALSE;
						if (options[optnum].msg_if_false)
							printf( "%s.\n", options[optnum].msg_if_false);
					}
					break;
				case OPT_INTEGER:
					if (argc <= 1)
						ProgError( "missing argument after \"%s\"", argv[0]);
					argv++;
					argc--;
					*((int *) (options[optnum].data_ptr)) = atoi( argv[0]);
					if (options[optnum].msg_if_true)
						printf( "%s: %d.\n", options[optnum].msg_if_true, atoi( argv[0]));
					break;
				case OPT_STRING:
					if (argc <= 1)
						ProgError( "missing argument after \"%s\"", argv[0]);
					argv++;
					argc--;
					*((char **) (options[optnum].data_ptr)) = argv[0];
					if (options[optnum].msg_if_true)
						printf( "%s: %s.\n", options[optnum].msg_if_true, argv[0]);
					break;
				case OPT_STRINGACC:
					if (argc <= 1)
						ProgError( "missing argument after \"%s\"", argv[0]);
					argv++;
					argc--;
					AppendItemToList( (char ***) options[optnum].data_ptr, argv[0]);
					if (options[optnum].msg_if_true)
						printf( "%s: %s.\n", options[optnum].msg_if_true, argv[0]);
					break;
				case OPT_STRINGLIST:
					if (argc <= 1)
						ProgError( "missing argument after \"%s\"", argv[0]);
					while (argc > 1 && argv[1][0] != '-' && argv[1][0] != '+')
					{
						argv++;
						argc--;
						AppendItemToList( (char ***) options[optnum].data_ptr, argv[0]);
						if (options[optnum].msg_if_true)
							printf( "%s: %s.\n", options[optnum].msg_if_true, argv[0]);
					}
					break;
				default:
					ProgError( "unknown option type (BUG!)");
				}
				break;
			}
		}
		if (options[optnum].opt_type == OPT_END)
			ProgError( "invalid argument: \"%s\"", argv[0]);
		argv++;
		argc--;
	}
}


/*
	output the program usage to the specified file
*/
void Usage( FILE *where)
{
	fprintf( where, "Usage:\n");
	fprintf( where, "sfinfo [+a|p] [-t <thing>] [-o <obj>] [-e <enum>]\n");
	fprintf( where, "   +a    Display all objects (also +allobj)\n");
	fprintf( where, "   +p    Display pain chance table (also +paintab)\n");
	fprintf( where, "   +m    Display monsters with pain chance & hit points (also +monsters)\n");
	fprintf( where, "   -w    Display data for specified weapon (also -weapon)\n");
	fprintf( where, "   -t    Display data for specified thing (also -thing)\n");
	fprintf( where, "   -o    Display data for specified object (also -object)\n");
	fprintf( where, "   -e    Display data for specified enum (also -enum)\n");
	fprintf( where, "Put a '+' instead of a '-' before boolean options to reverse their effect\n");
}

/*
	output the credits of the program to the specified file
*/
void Credits( FILE *where)
{
	fprintf( where, "SFINFO: StriFe INFO.c processor, ver %s\n", SFINFO_VERSION);
	fprintf( where, " By Frans P. de Vries <fpv@gamers.org>\n");
}


/*
	terminate the program reporting an error
*/
void ProgError( char *errstr, ...)
{
	va_list args;

	va_start( args, errstr);
	printf( "\nProgram Error: *** ");
	vprintf( errstr, args);
	printf( " ***\n");
	va_end( args);
	/* clean up & free space */
	exit( 1);
}

/*
	convert string to uppercase
*/
char *strupr( char *str)
{
	char *s = str;

	if (str != NULL)
		while (*s)
		{
			if (islower( *s))
				*s = toupper( *s);
			s++;
		}

	return s;
}

/* vim:set noexpandtab tabstop=2 softtabstop=2 shiftwidth=2: */
